/****************************************************************************
 *
 * bxcl1: Erste Version der Klassenbibliothek BXCL
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 5.1: Die Beispiel-Klassenbibliothek BXCL
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	bxtypes.h
 *
 ****************************************************************************/

#ifndef _bxtypes_h
#define _bxtypes_h

/* Fuer Object-IDs
 *  - kann spaeter als Klasse deklariert werden
 */
typedef long bxid;
static const bxid bxnoid = -1;

#endif  /* _bxtypes_h */


/****************************************************************************
 *
 *	bxObject.h
 *
 ****************************************************************************/

#ifndef _bxObject_h
#define _bxObject_h

// Standard-Header-Dateien einbinden
#include <iostream.h>
#include <stdlib.h>

// Hilfstypen
//#include "bxtypes.h"

/* Abstrakte Klasse bxObject
 *  - Wurzel einer Klassenbibliothek
 *  - definiert alle Standard-Funktionen fuer Objekte
 */
class bxObject {
  protected:
    static bxid maxId;    // aktuelle hoechste ID
    const  bxid objId;    // jeweilige Objekt-ID

  public:
    // Virtueller Destruktor
    virtual ~bxObject () {
    }

    // Klassen-Name
    virtual const char* className () const = 0;

    /* Test auf identische Gleichheit
     *  - wird von allen Klassen einfach geerbt
     */
    virtual bool isSame (const bxObject&) const;

    /* Test auf Wertgleichheit mit beliebigem Objekt
     *  - basiert auf Test auf Wertgleichheit mit einem
     *    Objekt der gleichen Klasse (operator==)
     */
    virtual bool isEqual (const bxObject&) const = 0;

    // Ausgabe
    virtual void printOn (ostream& = cout) const = 0;

  protected:
    // Default-Konstruktor
    bxObject ()
      : objId(++maxId) {    // ID initialisieren
    }

    // Copy-Konstruktor
    bxObject (const bxObject&)
      : objId(++maxId) {    // ID initialisieren
    }

    // Zuweisung
    void operator = (const bxObject&) {
                            // ID nicht zuweisen
    }
};

/* Ausgabeoperator "<<" fuer alle Klassen deklarieren
 */
ostream& operator << (ostream& strm, const bxObject& obj);

#endif  /* _bxObject_h */


/****************************************************************************
 *
 *	bxObject.cc / bxObject.cpp
 *
 ****************************************************************************/

/* Quelldatei zur abstrakten Klasse bxObject
 */

// Header-Datei der eigenen Klasse einbinden
//#include "bxObject.h"

/* statische Komponente maxId initialisieren
 */
bxid bxObject::maxId = 0;

/* operator <<
 *  - Ausgabeoperator fuer alle Objekte
 */
ostream& operator << (ostream& strm, const bxObject& obj)
{
    // objektspezifische Ausgabe-Funktion aufrufen
    obj.printOn (strm);

    // fuer Verkettung
    return strm;
}

/* isSame
 *  - testet auf Identitaet durch Vergleich der IDs
 *  - geht auch bei Mehrfachvererbung
 */
bool bxObject::isSame (const bxObject& obj) const
{
    return  objId == obj.objId;
}



/****************************************************************************
 *
 *	bxInt.h
 *
 ****************************************************************************/

#ifndef _bxInteger_h
#define _bxInteger_h

// Header-Datei der Basisklasse
//#include "bxObject.h"

/* Klasse bxInteger
 *  - abgeleitet von bxObject
 */
class bxInteger : public bxObject {

  protected:
    int theValue;                // Integer-Wert

  public:
    // Konstruktoren
    bxInteger (int i = 0) : theValue(i) {
    }
    // Wert abfragen
    virtual int value () const {
        return theValue;
    }

    // Wert aendern (neuen Wert zurueck)
    virtual int value (int i) {
        return theValue = i;
    }

    // Klassen-Name
    virtual const char* className () const;

    // Test auf Wertgleichheit mit beliebigem Objekt
    virtual bool isEqual (const bxObject&) const;

    // Test auf Wertgleichheit mit Objekt gleicher Klasse
    virtual bool operator== (const bxInteger&) const;
    virtual bool operator!= (const bxInteger&) const;

    // Ausgabe
    virtual void printOn (ostream& = cout) const;

    // Copy-Konstruktor, da die Basisklasse ihn definiert
    bxInteger (const bxInteger& i)
        : bxObject(i), theValue(i.theValue) {
    }

    // Zuweisung, da die Basisklasse sie definiert
    const bxInteger& operator= (const bxInteger& i) {
        bxObject::operator=(i);    // Zuweisung der Basisklasse
        theValue = i.theValue;     // Wert zuweisen
        return *this;              // Objekt zurueckliefern
    }

    // virtueller Destruktor
    virtual ~bxInteger () {
    }
};

#endif  /* _bxInteger_h */


/****************************************************************************
 *
 *	bxInt.cc / bxInt.cpp
 *
 ****************************************************************************/

/* Quelldatei zur Klasse bxInteger
 */

// Header-Datei der eigenen Klasse einbinden
//#include "bxInteger.h"

const char* bxInteger::className () const
{
    return "bxInteger";
}

bool bxInteger::isEqual (const bxObject& obj) const
{
    return  className() == obj.className() &&
            *this == (const bxInteger&)(obj);
}

bool bxInteger::operator== (const bxInteger& x) const
{
    return  theValue == x.theValue;
}

bool bxInteger::operator!= (const bxInteger& x) const
{
    return  theValue != x.theValue;
}
void bxInteger::printOn (ostream& strm) const
{
    strm << theValue;
}


/****************************************************************************
 *
 *	bxString.h
 *
 ****************************************************************************/

#ifndef _bxString_h
#define _bxString_h

// Header-Datei der Basisklasse
//#include "bxObject.h"

/* Klasse bxString
 *  - abgeleitet von bxObject
 */
class bxString : public bxObject {
  protected:
    char*     cstring;    // Zeichenfolge als dynamischer C-String
    unsigned  len;        // Anzahl Zeichen (ohne '\0')
    unsigned  size;       // Speicherplatzgroesse von "cstring"

  public:
    // Default- und char*-Konstruktor
    bxString (const char* = "");

    // Aufgrund dynamischer Komponenten:
    bxString (const bxString&);                   // Copy-Konstruktor
    virtual ~bxString ();                         // virtueller Destruktor
    const bxString& operator= (const bxString&);  // Zuweisung

    // aneinanderhaengen von Strings
    virtual bxString operator+ (const bxString&) const;

    // Klassen-Name
    virtual const char* className () const;

    // Test auf Wertgleichheit mit beliebigem Objekt
    virtual bool isEqual (const bxObject&) const;

    // Test auf Wertgleichheit mit Objekt gleicher Klasse
    virtual bool operator== (const bxString&) const;
    virtual bool operator!= (const bxString&) const;

    // Ausgabe
    virtual void printOn (ostream& = cout) const;

  protected:
    /* Konstruktor fuer uninitialisierten String bestimmter Laenge
     * (intern fuer operator+())
     */
    bxString (unsigned);
};

#endif  /* _bxString_h */


/****************************************************************************
 *
 *	bxString.cc / bxString.cpp
 *
 ****************************************************************************/

/* Quelldatei zur Klasse bxString
 */

// Header-Dateien einbinden
#include <string.h>
//#include "bxString.h"

/* Default- und char*-Konstruktor
 *  - Default fuer s: ""
 */
bxString::bxString (const char* s)
{
    len = strlen (s);             // Anzahl Zeichen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
    strcpy (cstring, s);          // Zeichen in Speicherplatz kopieren
}

/* Copy-Konstruktor
 */
bxString::bxString (const bxString& s)
    : bxObject(s)                 // Copy-Konstruktor der Basisklasse
{
    len = s.len;                  // Anzahl Zeichen uebernehmen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
    strcpy (cstring, s.cstring);  // Zeichen kopieren
}

/* Destruktor
 */
bxString::~bxString ()
{
    // Mit new angelegten Speicherplatz wieder freigeben
    delete [] cstring;
}

/* operator =
 *  - Zuweisung
 */
const bxString& bxString::operator= (const bxString& s)
{
    // Zuweisung eines Strings an sich selbst macht nichts
    if (isSame(s)) {
        return *this;                 // String zurueckliefern    
    }

    // Zuweisungen der Basisklasse durchfuehren
    bxObject::operator= (s);

    len = s.len;             // Anzahl Zeichen uebernehmen

    // falls Platz nicht reicht, vergroessern
    if (size < len + 1) {
        delete [] cstring;            // alten Speicherplatz freigeben
        size = len + 1;               // Anzahl Zeichen plus '\0'
        cstring = new char [size];    // Speicherplatz anlegen
    }

    strcpy (cstring, s.cstring);      // Zeichen kopieren

    return *this;            // geaenderten String zurueckliefern    
}

/* operator +
 *  - haengt zwei Strings hintereinander
 */
bxString bxString::operator+ (const bxString& s) const
{
    // uninitialisierten Summen-String erzeugen
    bxString sum (len + s.len + 1);
 
    // Summen-Zeichenfolge darin initialisieren
    strcpy (sum.cstring, cstring);
    strcat (sum.cstring, s.cstring);

    // Kopie von Summen-String zurueck
    return sum;
}

/* Konstruktor fuer uninitialisierten String bestimmter Laenge
 *  - intern fuer operator+
 */
bxString::bxString (unsigned l)
{
    len = l;                      // Anzahl Zeichen uebernehmen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
}

const char* bxString::className () const
{
    return "bxString";
}

bool bxString::isEqual (const bxObject& obj) const
{
    return  className() == obj.className() &&
            *this == (const bxString&)(obj);
}

bool bxString::operator== (const bxString& x) const
{
    return  (strcmp(cstring,x.cstring) == 0);
}

bool bxString::operator!= (const bxString& x) const
{
    return  (strcmp(cstring,x.cstring) != 0);
}

void bxString::printOn (ostream& strm) const
{
    // Zeichenfolge einfach ausgeben
    strm << cstring;
}


/****************************************************************************
 *
 *	bxcl1.cc / bxcl1.cpp
 *
 ****************************************************************************/

// Header-Dateien
#include <iostream.h>
//#include "bxInteger.h"
//#include "bxString.h"

// Vorwaerts-Deklaration
void vergleiche (bxObject&, bxObject&);

main()
{
    bxString s1, s2, s3;
    bxInteger i1;

    s1 = s2 = "hallo";
    i1 = 17;

    vergleiche (s1, s2);      // gleicher Wert
    vergleiche (s1, s3);      // verschiedener Wert
    vergleiche (s1, i1);      // verschiedene Klasse
    vergleiche (s1, s1);      // identisch
}

void vergleiche (bxObject& o1, bxObject& o2)
{
    if (o1.isSame(o2)) {
        cout << "die beiden Objekte sind identisch (Wert: "
             << o1 << ')' << endl;
    }
    else if (o1.isEqual(o2)) {
        cout << "die beiden Objekte haben den gleichen Wert: "
             << o1 << endl;
    }
    else if (o1.className() == o2.className()) {
        cout << "die beiden Objekte haben verschiedene Werte: "
             << o1 << " und " << o2 << endl;
    }
    else {
        cout << "die beiden Objekte haben verschiedene Klassen ("
             << o1.className() << " und " << o2.className() << ")\n"
             << " und natuerlich auch verschiedene Werte: "
             << o1 << " und " << o2 << endl;
    }
}

