/****************************************************************************
 *
 * Person1: Klasse Person (verwendet Klasse String)
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 3.11: Klassen verwenden Klassen
 *    Abschnitt 3.12: Statische Komponenten und Hilfstypen
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	String.h
 *
 ****************************************************************************/

#ifndef _String_h
#define _String_h

// Header-Datei fuer I/O
#include <iostream.h>

class String {

  private:
    char*    cstring;    // Zeichenfolge als dynamischer C-String
    unsigned len;        // Anzahl Zeichen (ohne '\0')
    unsigned size;       // Speicherplatzgroesse von "cstring"

  public:
    // Default- und char*-Konstruktor
    String (const char* = "");

    // Vergleichen von Strings
    friend bool operator== (const String&, const String&);
    friend bool operator!= (const String&, const String&);

    // Hintereinanderhaengen von Strings
    friend String operator+ (const String&, const String&);

    // Ausgabe mit Streams
    void printOn (ostream& = cout) const;

    // Eingabe mit Streams
    void scanFrom (istream& = cin);

    // Aufgrund dynamischer Komponenten
    String (const String&);                   // Copy-Konstruktor
    const String& operator= (const String&);  // Zuweisung
    ~String ();                               // Destruktor
  private:
    /* Konstruktor fuer uninitialisierten String bestimmter Laenge
     * (intern fuer operator+())
     */
    String (unsigned);
};

// Standard-Ausgabeoperator
inline ostream& operator<< (ostream& strm, const String& s)
{
    s.printOn (strm);    // String auf Stream ausgeben
    return strm;         // Stream zurueckliefern
}

// Standard-Eingabeoperator
inline istream& operator>> (istream& strm, String& s)
{
    s.scanFrom (strm);   // String von Stream einlesen
    return strm;         // Stream zurueckliefern
}

/* Operator !=
 *  - als Umsetzung auf Operator == inline implementiert
 */
inline bool operator!= (const String& s1, const String& s2) {
    return !(s1==s2);
}

#endif  /* _String_h */


/****************************************************************************
 *
 *	String.cc / String.cpp
 *
 ****************************************************************************/

// Header-Datei der eigenen Klasse
//#include "String.h"

// C-Header-Dateien fuer String-Funktionen
#include <string.h>
#include <ctype.h>

/* Konstruktor aus C-String (char*)
 *  - Default fuer s: ""
 */

String::String (const char* s)
{
    len = strlen (s);             // Anzahl Zeichen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
    strcpy (cstring, s);          // Zeichen in Speicherplatz kopieren
}

/* Copy-Konstruktor
 */
String::String (const String& s)
{
    len = s.len;                    // Anzahl Zeichen uebernehmen
    size = len + 1;                 // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];      // Speicherplatz anlegen
    strcpy (cstring, s.cstring);    // Zeichen kopieren
}

/* Destruktor
 */
String::~String ()
{
    // Mit new angelegten Speicherplatz wieder freigeben
    delete [] cstring;
}

/* operator =
 *  - Zuweisung
 */
const String& String::operator= (const String& s)
{
    // Zuweisung eines Strings an sich selbst macht nichts
    if (this == &s) {
        return *this;                 // String zurueckliefern    
    }

    len = s.len;             // Anzahl Zeichen uebernehmen

    // falls Platz nicht reicht, vergroessern
    if (size < len + 1) {
        delete [] cstring;            // alten Speicherplatz freigeben
        size = len + 1;               // Anzahl Zeichen plus '\0'
        cstring = new char [size];    // Speicherplatz anlegen
    }

    strcpy (cstring, s.cstring);      // Zeichen kopieren

    return *this;            // geaenderten String zurueckliefern    
}

/* operator ==
 *  - vergleicht zwei Strings
 *  - globale Friend-Funktion, damit automatische
 *    Typumwandlung des ersten Operanden moeglich ist
 */
bool operator== (const String& s1, const String& s2)
{
    return strcmp(s1.cstring,s2.cstring) == 0;
}

/* operator +
 *  - haengt zwei Strings hintereinander
 *  - globale Friend-Funktion, damit automatische
 *    Typumwandlung des ersten Operanden moeglich ist
 */
String operator+ (const String& s1, const String& s2)
{
    // uninitialisierten Summen-String erzeugen
    String sum (s1.len + s2.len + 1);

    // Summen-Zeichenfolge darin initialisieren
    strcpy (sum.cstring, s1.cstring);
    strcat (sum.cstring, s2.cstring);

    // Kopie von Summen-String zurueck
    return sum;
}

/* Konstruktor fuer uninitialisierten String bestimmter Laenge
 *  - intern fuer operator+
 */
String::String (unsigned l)
{
    len = l;                      // Anzahl Zeichen uebernehmen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
}

/* Ausgabe auf Stream
 */
void String::printOn (ostream& strm) const
{
    // Zeichenfolge einfach ausgeben
    strm << cstring;
}

void String::scanFrom (istream& strm)
{
    char c;

    len = 0;         // zunaechst ist der gelesene String leer

    strm >> ws;      // fuehrende Trennzeichen ueberlesen

    /* solange Input-Stream "strm" nach Einlesen
     * eines Zeichens in "c" in Ordnung ist
     */
    while (strm.get(c)) {       // >> wuerde Trennzeichen ueberlesen

        /* Newline schliesst Stringeingabe ab
         *  - '\n' => RETURN
         */
        if (c == '\n') {
            cstring [len] = '\0';
            return;
        }

        /* falls der Platz nicht mehr ausreicht,
         * muss mehr geschaffen werden
         */
        if (len+1 >= size) {
            char* tmp = cstring;       // Zeiger auf alten Speicherplatz
            size += 80;                // 80 Zeichen mehr Platz
            cstring = new char[size];  // neuen Speicherplatz anlegen
            strcpy (cstring, tmp);     // Zeichen kopieren
            delete [] tmp;             // alten Speicherplatz freigeben
        }

        // neues Zeichen eintragen
        cstring [len] = c;
        len ++;
    }

    /* Einlese-Ende durch Fehler oder EOF
     *  - Stringende-Kennzeichen eintragen
     */
    cstring [len] = '\0';
}



/****************************************************************************
 *
 *	Person.h
 *
 ****************************************************************************/

#ifndef _Person_h
#define _Person_h

// Header-Dateien einbinden
#include <iostream.h>
#include <stdlib.h>
//#include "String.h"

class Person {
  // statische Klassenkomponenten
  private:
    static long personenMaxID;   // hoechste ID aller Personen
    static long personenAnzahl;  // aktuelle Anzahl aller Personen
  public:
    // aktuelle Anzahl aller Personen liefern
    static long anzahl () {
        return personenAnzahl;
    }

  /* NEU: spezielle Datentypen
   */
  public:
    // Aufzaehlungstyp fuer die Anrede
    enum Anrede { herr, frau, firma, leer };
  
  // nicht-statische Klassenkomponenten
  private:
    Anrede     anr;          // NEU: Anrede (kann auch leer sein)
    String     vname;        // Vorname der Person
    String     nname;        // Nachname der Person
    const long pid;          // eindeutige ID der Person

  public:
    /* Konstruktor aus Nachname, optional Vorname
     * und (NEU:) optional Anrede
     */
    Person (const String&, const String& = "", const Anrede& = leer);

    // Copy-Konstruktor
    Person (const Person&);

    // Destruktor
    ~Person ();

    // Zuweisung
    const Person& operator = (const Person&);

    // Abfragen von Eigenschaften
    const String& nachname () const {      // Nachname liefern
        return nname;
    }
    const String& vorname () const {       // Vorname liefern
        return vname;
    }
    const long id () const {               // ID liefern
        return pid;
    }
    const Anrede& anrede () const {        // NEU: Anrede liefern
        return anr;                        // (eigener Aufzaehlungstyp)
    }

    // Vergleichen
    bool operator == (const Person& p) const {
        return vname == p.vname && nname == p.nname;
    }
    bool operator != (const Person& p) const {
        return vname != p.vname || nname != p.nname;
    }
};

#endif  /* _Person_h */


/****************************************************************************
 *
 *	Person.cc / Person.cpp
 *
 ****************************************************************************/

// Header-Datei der Klasse
//#include "Person.h"

/* Statische Klassen-Komponenten initialisieren
 */
long Person::personenMaxID = 0;
long Person::personenAnzahl = 0;

/* Konstruktor aus Nachname und Vorname
 * (Default fuer Vorname: "", Default fuer Anrede: leer)
 *
 *  - Vor- und Nachname, sowie Anrede werden mit Initialisierungsliste
 *    direkt an die Konstruktoren fuer nname und vname uebergeben
 *  - Die ID wird ebenfalls direkt initialisiert
 */
Person::Person (const String& nn, const String& vn, const Anrede& a)
  : nname(nn), vname(vn), anr(a), pid(++personenMaxID)
{
    // Anzahl der existierenden Personen erhoehen
    personenAnzahl++;
}

/* Copy-Konstruktor
 */
Person::Person (const Person& p)
  : nname(p.nname), vname(p.vname), pid(++personenMaxID)
{
    // Anzahl der existierenden Personen erhoehen
    personenAnzahl++;
}

/* Destruktor
 */
Person::~Person ()
{
    // Anzahl der existierenden Personen herabsetzen
    personenAnzahl--;
}

/* Zuweisung
 */
const Person& Person::operator = (const Person& p)
{
    if (this == &p) {
        return *this;
    }

    // alles ausser ID zuweisen
    nname = p.nname;
    vname = p.vname;
    anr = p.anr;

    return *this;
}


/****************************************************************************
 *
 *	ptest1.cc / ptest1.cpp
 *
 ****************************************************************************/

#include <iostream.h>
//#include "Person.h"

void printPerson (const Person&);


main ()
{
    /* Array von 4 Personen
     *  - ohne Default-Konstruktor muss es Werte zur Initialisierung geben,
     *    weshalb es statisch sein muss
     */
    static Person typen [4] = { Person("****"), Person("****"), Person("****"), Person("****") };

    /* Vier mehr oder weniger normale Personen(-Konstanten)
     */
    const Person nico ("Josuttis", "Nicolai", Person::herr);
    const Person bredex ("BREDEX GmbH", "", Person::firma);
    const Person spock ("Spock", "Mister");
    const Person alf ("Alf");

    cout << "\nPersonen:" << endl;
    printPerson (nico);    
    printPerson (bredex);    
    printPerson (spock);    
    printPerson (alf);    

    typen[0] = nico;
    typen[1] = bredex;
    typen[2] = spock;
    typen[3] = alf;

    cout << "\nals Typen:" << endl;
    for (int i=0; i<4; i++) {
        printPerson (typen[i]);    
    }
}

void printPerson (const Person& p)
{
    switch (p.anrede()) {
    	case Person::herr:
    	    cout << "Herr ";
    	    break;
    	case Person::frau:
    	    cout << "Frau ";
    	    break;
    	case Person::firma:
    	    cout << "Firma ";
    	    break;
    }
    if (p.vorname() != "") {
    	cout << p.vorname() << ' ';
    }
    if (p.nachname() != "") {
    	cout << p.nachname() << ' ';
    }
    cout << "(ID: " << p.id() << ')' << endl;
}

